//
// Created by fabio on 9/13/24.
//

#include "drip.h"
#include "utils.h"

#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <errno.h>

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

static size_t wrapInEndorsementStructure(DRIP_SAM_Data_Format *payload, unsigned char *evidence, size_t len, DRIP_HostIdentity *drone_hi);
static int prepareASTMAuthMessage(ODID_Auth_encoded *encoded, DRIP_SAM_Data_Format *payload, size_t len, short *pages_out);
static void (*sign_data)(DRIP_HostIdentity*, const unsigned char*, size_t, unsigned char*, size_t*) = NULL;
static time_t (*astm_time)() = NULL;
static DRIP_Core_Config *config = NULL;

void DRIP_SetSignDataFun(void (*fun)(DRIP_HostIdentity*, const unsigned char*, size_t, unsigned char*, size_t*)) {
    sign_data = fun;
}

void DRIP_SetASTMDateTimeFun(time_t (*fun)()) {
    astm_time = fun;
}

void DRIP_SetConfig(DRIP_Core_Config *conf) {
    config = conf;
}

int encodeDripLink(ODID_Auth_encoded *encoded, const char* endorsement, short *pages_out) {
    size_t bytes_len;
    unsigned char endorsement_bytes[DRIP_ENDORSEMENT_LEN];
    hex_string_to_bytes(endorsement, endorsement_bytes, &bytes_len);

    //Building DRIP_LINK payload
    DRIP_SAM_Data_Format payload;
    payload.sam_type = DRIP_LINK;
    memcpy(payload.sam_auth_data, endorsement_bytes, bytes_len);

    return prepareASTMAuthMessage(encoded, &payload, bytes_len, pages_out);
}

int encodeDripManifest(
    ODID_Auth_encoded *encoded,
    DRIP_HostIdentity *drone_hi,
    ODID_Message_encoded *messages,
    const size_t cnt,
    const unsigned char *previous_manifest_hash,
    const unsigned char *drip_link_be_hda_ua_hash,
    unsigned char *next_manifest_hash_out,
    short *pages_out
) {

    if(next_manifest_hash_out == NULL) {
        perror("next_manifest_hash_out not allocated");
        return ODID_FAIL;
    }

    if(cnt > DRIP_MAX_MANIFEST_MSGS) {
        perror("Tried to broadcast manifest with too many messages");
        return ODID_FAIL;
    }

    unsigned char *msg_hashes = malloc(DRIP_HASH_LEN*cnt); // a message hash will be 8 byte long
    for(int i = 0; i < cnt; i++) {
        cshake128_wrapper(
            msg_hashes+(i*DRIP_HASH_LEN),
            (uint8_t*)messages+(i*sizeof(ODID_Message_encoded)),
            sizeof(ODID_Message_encoded)
        );
    }

    //Building DRIP_MANIFEST payload
    DRIP_SAM_Data_Format payload;
    payload.sam_type = DRIP_MANIFEST;

    //evidence consist of previous and current manifest hash (initially NULL) and ASTM messages hashes
    const unsigned int evidence_len = DRIP_HASH_LEN*(3+cnt);
    unsigned char *evidence = malloc(evidence_len);

    //set previous manifest_hash
    memcpy(evidence, previous_manifest_hash, DRIP_HASH_LEN);

    //temporary set current manifest hash to null (check 4.4.3 RFC9575)
    memset(evidence+DRIP_HASH_LEN, 0x00, DRIP_HASH_LEN);

    //copy drip_link_be_hda_ua to evidence as per RFC9575
    memcpy(evidence+(DRIP_HASH_LEN*2), drip_link_be_hda_ua_hash, DRIP_HASH_LEN);

    memcpy(evidence+(DRIP_HASH_LEN*3), msg_hashes, DRIP_HASH_LEN*cnt);
    free(msg_hashes);

    unsigned char current_manifest_hash[DRIP_HASH_LEN];
    cshake128_wrapper(current_manifest_hash, evidence, DRIP_HASH_LEN);

    //setting the REAL current manifest hash
    memcpy(evidence+DRIP_HASH_LEN, current_manifest_hash, DRIP_HASH_LEN);

    size_t payload_len = wrapInEndorsementStructure(&payload, evidence, evidence_len, drone_hi);
    free(evidence);

    //provide current_hash to user
    memcpy(next_manifest_hash_out, current_manifest_hash, DRIP_HASH_LEN);

    return prepareASTMAuthMessage(encoded, &payload, payload_len, pages_out);
}

int encodeDripWrapper(
    ODID_Auth_encoded *encoded,
    DRIP_HostIdentity *drone_hi,
    ODID_Message_encoded *messages,
    size_t cnt,
    short *pages_out
) {
    if(cnt > DRIP_MAX_MANIFEST_MSGS) {
        perror("Tried to broadcast wrapper with too many messages");
        return ODID_FAIL;
    }

    for(int i = 0; i < cnt; i++) {
        //only message 0x0 0x1 0x3 0x4 0x5 are allowed
        if(messages[i].rawData[1] == ODID_MESSAGETYPE_AUTH ||
            messages[i].rawData[1] == ODID_MESSAGETYPE_INVALID ||
            messages[i].rawData[1] == ODID_MESSAGETYPE_PACKED) {
            perror("Can't encode DRIP Wrapper with Authentication Message");
            return ODID_FAIL;
        }
    }

    //Building DRIP_WRAPPER payload
    DRIP_SAM_Data_Format payload;
    payload.sam_type = DRIP_WRAPPER;

    size_t evidence_len = sizeof(ODID_Message_encoded) * cnt;
    unsigned char evidence[evidence_len];

    //sort messages by type (0x0, 0x1, 0x3, 0x4, 0x5)
    //messages have to be sorted by type according to RFC9575
    qsort(messages, cnt, sizeof(ODID_Message_encoded), compareODIDMessages);

    memcpy(evidence, messages, evidence_len);
    size_t payload_len = wrapInEndorsementStructure(&payload, evidence, evidence_len,drone_hi);

    return prepareASTMAuthMessage(encoded, &payload, payload_len, pages_out);
}

int encodeDripFrame(
    ODID_Auth_encoded *encoded,
    DRIP_HostIdentity *drone_hi,
    const DRIP_Frame_data *data,
    const size_t evidence_len,
    short *pages_out
) {
    //Building DRIP_FRAME payload
    DRIP_SAM_Data_Format payload;
    payload.sam_type = DRIP_FRAME;

    size_t payload_len = wrapInEndorsementStructure(&payload, (unsigned char*)data, evidence_len, drone_hi);
    return prepareASTMAuthMessage(encoded, &payload, payload_len, pages_out);
}

static size_t wrapInEndorsementStructure(DRIP_SAM_Data_Format *payload, unsigned char *evidence, size_t len, DRIP_HostIdentity *drone_hi) {
    int32_t time = astm_time();

    //2.2 RFC9575
    int32_t time_vna = time + 120;

    unsigned char VNB_timestamp[DRIP_VNX_TIMESTAMP_LEN];
    unsigned char VNA_timestamp[DRIP_VNX_TIMESTAMP_LEN];

    //must fit in 4 bytes, but time_t is long int (at least on x86)
    memcpy(VNB_timestamp, (unsigned char*)&time, DRIP_VNX_TIMESTAMP_LEN);
    memcpy(VNA_timestamp, (unsigned char*)&time_vna, DRIP_VNX_TIMESTAMP_LEN);

    memcpy(payload->sam_auth_data, VNB_timestamp, DRIP_VNX_TIMESTAMP_LEN);
    memcpy(payload->sam_auth_data+DRIP_VNX_TIMESTAMP_LEN, VNA_timestamp, DRIP_VNX_TIMESTAMP_LEN);
    memcpy(payload->sam_auth_data+(2*DRIP_VNX_TIMESTAMP_LEN), evidence, len);
    memcpy(payload->sam_auth_data+(2*DRIP_VNX_TIMESTAMP_LEN+len), drone_hi->hhit, DRIP_HHIT_LEN);

    size_t unsigned_message_len = 2*DRIP_VNX_TIMESTAMP_LEN + len + DRIP_HHIT_LEN;
    size_t signed_message_len;
    unsigned char signed_message[DRIP_UA_SIGNATURE_LEN];

    if(sign_data != NULL)
        sign_data(drone_hi, payload->sam_auth_data, unsigned_message_len, signed_message, &signed_message_len);
    else
        perror("sign_data function not set");

    memcpy(payload->sam_auth_data+unsigned_message_len, signed_message, signed_message_len);

    return unsigned_message_len + DRIP_UA_SIGNATURE_LEN;
}

static int prepareASTMAuthMessage(ODID_Auth_encoded *encoded, DRIP_SAM_Data_Format *payload, size_t len,  short *pages_out) {
    size_t total_len = len + 1; //1 byte for SAM_TYPE

    if(total_len > DRIP_MAX_SAM_LEN) {
        perror("Tried to broadcast an auth message that is too long");
        return ODID_FAIL;
    }

    if(config == NULL) {
        perror("DRIP config not set");
        return ODID_FAIL;
    }

    size_t first_page_len = MIN(total_len, ODID_AUTH_PAGE_ZERO_DATA_SIZE);
    size_t remaining_len = MAX(0, total_len - ODID_AUTH_PAGE_ZERO_DATA_SIZE);
    size_t last_page_index = ceil((float)remaining_len / ODID_AUTH_PAGE_NONZERO_DATA_SIZE);
    size_t last_page_len_used_bytes = remaining_len % ODID_AUTH_PAGE_NONZERO_DATA_SIZE;

    uint8_t additional_needed_pages = 0;

    if(config->legacy_transport) {
        //For FEC either 2 or 1 additional pages are needed (RFC9575 5.3)
        additional_needed_pages = remaining_len % ODID_AUTH_PAGE_NONZERO_DATA_SIZE == 0 ? 2 : 1;
    }

    ODID_Auth_data auth_data[1 + last_page_index + additional_needed_pages];

    for(int i = 0; i <= last_page_index; i++) {
        odid_initAuthData(&auth_data[i]);

        // Set SAM type to 0x5
        auth_data[i].AuthType = ODID_AUTH_SPECIFIC_AUTHENTICATION;
        auth_data[i].DataPage = i;

        //first page only
        if(i == 0) {
            auth_data[i].LastPageIndex = last_page_index + additional_needed_pages;
            auth_data[i].Length = total_len;
            auth_data[i].Timestamp = astm_time();
            memcpy(auth_data[i].AuthData, (unsigned char *)payload, first_page_len);
        }
        else
            memcpy(
                auth_data[i].AuthData,
                ((unsigned char *)payload) + first_page_len + (i-1)*ODID_AUTH_PAGE_NONZERO_DATA_SIZE,
                i == last_page_index ? last_page_len_used_bytes : ODID_AUTH_PAGE_NONZERO_DATA_SIZE
            );

        //encode all pages beside last which might need FEC "Additional Data Length byte"
        if(i != last_page_index) {
            const int ret = encodeAuthMessage(&encoded[i], &auth_data[i]);

            //very verbose, very cutesy, very demure. Compiler's gonna eat it and leave no crumbs.
            if(ret == ODID_FAIL)
                return ODID_FAIL;
            else continue;
        }
    }

    //handle last page separately

    int ret = ODID_SUCCESS;
    if(config->legacy_transport) {
        //FEC is needed for legacy transport as per RFC9575
        //copy ADL to last page if any space left
        if(additional_needed_pages == 1) {
            memset(auth_data[last_page_index].AuthData+last_page_len_used_bytes, ODID_AUTH_PAGE_NONZERO_DATA_SIZE, sizeof(uint8_t));
            //null padding for the rest of the page not needed (already done by odid_initAuthData)
            ret &= encodeAuthMessage(&encoded[last_page_index], &auth_data[last_page_index]);
        }
        else
        {
            //space not left. Copy ADL in new page
            odid_initAuthData(&auth_data[last_page_index+1]);

            auth_data[last_page_index+1].AuthType = ODID_AUTH_SPECIFIC_AUTHENTICATION;
            auth_data[last_page_index+1].DataPage = last_page_index+1;

            //null padding for the rest of the page not needed (already done by odid_initAuthData)
            auth_data[last_page_index+1].AuthData[0] = ODID_AUTH_PAGE_NONZERO_DATA_SIZE;
            ret &= encodeAuthMessage(&encoded[last_page_index+1], &auth_data[last_page_index+1]);
        }

        //calculate forward error correction in final page
        //FEC needs to be calculated on the already encoded data because of the way opendroneid api
        //aligns the data in the structs.
        uint8_t fec[ODID_AUTH_PAGE_NONZERO_DATA_SIZE] = {0};
        for(int i = 0; i<=last_page_index; i++) {
            for(int j = 0; j<ODID_AUTH_PAGE_NONZERO_DATA_SIZE; j++) {
                //j+2 because first byte is ASTM [MessageType][ProtoVersion]
                //todo    second byte is [DataPage][AuthType], but skip it might be not up to standard
                fec[j] ^= ((uint8_t*)&encoded[i])[j+2];
            }
        }

        //additional_needed_pages is either 1 or 2
        last_page_index = last_page_index+additional_needed_pages;
        odid_initAuthData(&auth_data[last_page_index]);

        // Set SAM type to 0x5
        auth_data[last_page_index].AuthType = ODID_AUTH_SPECIFIC_AUTHENTICATION;
        auth_data[last_page_index].DataPage = last_page_index;
        memcpy(auth_data[last_page_index].AuthData, fec, ODID_AUTH_PAGE_NONZERO_DATA_SIZE);

        ret &= encodeAuthMessage(&encoded[last_page_index], &auth_data[last_page_index]);
    }
    else
        ret &= encodeAuthMessage(&encoded[last_page_index], &auth_data[last_page_index]);


    *pages_out = last_page_index + 1;
    return ret;
}
