//
// Created by fabio on 9/14/24.
//

#include "drip.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//keccak for cshake128
#include "utils.h"


#define MFR_CODE "8653"
#define LENGTH_CODE "F"

static int verify_hhit(const unsigned char* hhit, const unsigned char* pub_key);
static void generate_serial_number(const unsigned char* hhit, char* serial_out);

DRIP_HostIdentity * DRIP_HostIdentityFromBytes(const unsigned char *priv_key, const unsigned char *pub_key,
    const unsigned char *hhit) {
    DRIP_HostIdentity* identity = malloc(sizeof(DRIP_HostIdentity));
    if (!identity) return NULL;

    //assume key are ED25519
    memcpy(identity->priv_key, priv_key, DRIP_ED25519_KEY_LEN);
    memcpy(identity->pub_key, pub_key, DRIP_ED25519_KEY_LEN);
    memcpy(identity->hhit, hhit, DRIP_HHIT_LEN);

    // Generate serial number
    generate_serial_number(identity->hhit, identity->serial_number);

    // Verify HHIT
    identity->is_hhit_verified = verify_hhit(identity->hhit, identity->pub_key);
    if (!identity->is_hhit_verified) {
        printf("The HHIT hash does not match.\n");
    } else {
        printf("The HHIT hash is verified.\n");
    }

    return identity;
}


static char* encode_MFR_SN(const char* MFR_SN_binary) {
    size_t binary_len = strlen(MFR_SN_binary);
    size_t encoded_len = binary_len / 5;  // Each 5 bits maps to 1 character

    const char encoding_dictionary[] = {
        '0', '1', '2', '3', '4', '5', '6', '7',
        '8', '9', 'A', 'B', 'C', 'D', 'E', 'F',
        'G', 'H', 'J', 'K', 'L', 'M', 'N', 'P',
        'Q', 'R', 'T', 'U', 'V', 'W', 'X', 'Y'
    };

    // Allocate memory for the encoded string (+1 for null terminator)
    char* MFR_SN_encoded = malloc(encoded_len + 1);
    if (!MFR_SN_encoded) {
        return NULL;
    }

    // Loop over the binary string in chunks of 5 bits
    for (size_t i = 0; i < binary_len; i += 5) {
        // Extract a 5-bit chunk
        char bit_sequence[6];  // 5 bits + null terminator
        strncpy(bit_sequence, MFR_SN_binary + i, 5);
        bit_sequence[5] = '\0';  // Null-terminate the bit sequence

        // Convert the bit sequence to an integer
        int index = strtol(bit_sequence, NULL, 2);  // Base 2 (binary)

        // Map the integer to the corresponding character in the encoding dictionary
        MFR_SN_encoded[i / 5] = encoding_dictionary[index];
    }

    MFR_SN_encoded[encoded_len] = '\0';  // Null-terminate the encoded string

    return MFR_SN_encoded;
}

// Serial number generation function
static void generate_serial_number(const unsigned char* hhit, char* serial_out) {
    /*
    Serial Number = MFR Code | Length Code | MFR SN
    MFR Code = ICAO MFR Code
    Length Code = 0xF
    MFR SN = HHIT Suite ID | ORCHID hash left-padded with 3 bits of zeros and encoded into 15 characters
    */

    // Extract Suite ID and ORCHID hash
    char hhit_suite_id = hhit[7];  // Assuming it's in the 8th byte
    char orchid_hash[8];
    memcpy(orchid_hash, hhit + 8, 8);

    // Encode Suite ID and ORCHID hash into MFR SN

    unsigned char combined[9];
    combined[0] = hhit_suite_id;
    memcpy(&combined[1], orchid_hash, 8);  // Copy ORCHID_hash after the first byte

    // Convert the concatenated byte array into an integer
    unsigned long long int MFR_SN_value = from_bytes(combined, 9);

    // Convert to a 75-bit binary string
    char* MFR_SN_binary = to_binary_string_padded(MFR_SN_value, 75);
    if(MFR_SN_binary != NULL) {
        char* MFR_SN_encoded = encode_MFR_SN(MFR_SN_binary);
        free(MFR_SN_binary);

        // Concatenate MFR code, Length code, and encoded MFR SN
        snprintf(serial_out, DRIP_SERIAL_NUMBER_LEN, "%s%s%s", MFR_CODE, LENGTH_CODE, MFR_SN_encoded);
    }
}

// Function to verify HHIT
static int verify_hhit(const unsigned char *hhit, const unsigned char *pkey) {
    unsigned char hash[8];
    unsigned char hhit_part[8];

    //assume is ED25519
    size_t pub_key_len = DRIP_ED25519_KEY_LEN;

    unsigned char combined[8 + pub_key_len];
    memcpy(combined, hhit, 8);
    memcpy(combined + 8, pkey, pub_key_len);

    // Compute hash of the combined data
    cshake128_wrapper(hash, combined, sizeof(combined));

    // Extract the part of HHIT to compare with the hash
    memcpy(hhit_part, hhit + 8, 8);

    // Compare the computed hash with the extracted part of HHIT
    return memcmp(hash, hhit_part, 8) == 0;
}

