//
// Created by fabio on 9/26/24.
//

#include "utils.h"
#include <drip.h>

#include <openssl/ec.h>
#include <openssl/pem.h>
#include <openssl/err.h>
#include <openssl/evp.h>

DRIP_HostIdentity* DRIP_HostIdentityFromFile(const char* identities_path, const char* identity_name) {
    char pub_key_path[512];
    char priv_key_path[512];
    char det_path[512];
    EVP_PKEY* pub_key;
    EVP_PKEY* priv_key;

    snprintf(pub_key_path, sizeof(pub_key_path), "%s/%s/%spub.pem", identities_path, identity_name, identity_name);
    snprintf(priv_key_path, sizeof(priv_key_path), "%s/%s/%sprv.pem", identities_path, identity_name, identity_name);
    snprintf(det_path, sizeof(det_path), "%s/%s/%s.det", identities_path, identity_name, identity_name);


    // Load the private key from PEM file
    FILE* priv_key_file = fopen(priv_key_path, "r");
    if (!priv_key_file) {
        perror("Error opening PEM file");
        return NULL;
    }
    if(!((priv_key = PEM_read_PrivateKey(priv_key_file, NULL, NULL, NULL)))) {
        ERR_print_errors_fp(stderr);
        fclose(priv_key_file);
        exit(EXIT_FAILURE);
    }
    fclose(priv_key_file);


    // Load the public key from PEM file
    FILE* pub_key_file = fopen(pub_key_path, "r");
    if (!pub_key_file) {
        perror("Error opening pub file");
        EVP_PKEY_free(priv_key);
        return NULL;
    }
    if(!((pub_key = PEM_read_PUBKEY(pub_key_file, NULL, NULL, NULL)))) {
        ERR_print_errors_fp(stderr);
        fclose(pub_key_file);
        EVP_PKEY_free(priv_key);
        exit(EXIT_FAILURE);
    }
    fclose(pub_key_file);



    // Load DET
    FILE* det_file = fopen(det_path, "r");
    if (!det_file) {
        perror("Error opening DET file");
        EVP_PKEY_free(pub_key);
        EVP_PKEY_free(priv_key);
        return NULL;
    }
    char hhit_str[DRIP_HHIT_LEN*2 + 1];
    char hhit[DRIP_HHIT_LEN];

    if(fgets(hhit_str, sizeof(unsigned char)*(DRIP_HHIT_LEN*2 + 1), det_file) == NULL)
    {
        perror("Error reading DET file");
        EVP_PKEY_free(pub_key);
        EVP_PKEY_free(priv_key);
        return NULL;
    }
    fclose(det_file);

    size_t bytes_len;
    hex_string_to_bytes(hhit_str, hhit, &bytes_len);

    size_t pub_key_len;
    size_t priv_key_len;

    if (priv_key == NULL) {
        printf("EVP_PKEY is NULL\n");
        return NULL;
    }

    int key_type = EVP_PKEY_base_id(priv_key);
    if (key_type == EVP_PKEY_ED25519) {
        printf("The key is of type ED25519.\n");
    } else {
        printf("The key is not ED25519, it is of type: %d\n", key_type);
    }

    //extract public and private key
    // Ensure pub_key_len has the correct value before calling this function
    EVP_PKEY_get_raw_public_key(priv_key, NULL, &pub_key_len);
    EVP_PKEY_get_raw_private_key(priv_key, NULL, &priv_key_len);

    unsigned char priv_key_bytes[priv_key_len];
    unsigned char pub_key_bytes[pub_key_len];

    if(EVP_PKEY_get_raw_public_key(priv_key, pub_key_bytes, &pub_key_len) <= 0)
        perror("Error reading private key");

    if(EVP_PKEY_get_raw_private_key(priv_key, priv_key_bytes, &priv_key_len) <= 0)
        perror("Error reading private key");

    return DRIP_HostIdentityFromBytes(priv_key_bytes, pub_key_bytes, hhit);
}

// Signing function
void sign_data(
    DRIP_HostIdentity* hi,
    const unsigned char* data,
    const size_t data_len,
    unsigned char* signed_data,
    size_t* signed_data_len
) {

    // Create a signing context
    EVP_MD_CTX *mdctx = EVP_MD_CTX_new();
    if (mdctx == NULL) {
        fprintf(stderr, "Error creating context\n");
        ERR_print_errors_fp(stderr);
    }

    EVP_PKEY *pkey = EVP_PKEY_new_raw_private_key(EVP_PKEY_ED25519, NULL, hi->priv_key, DRIP_ED25519_KEY_LEN);

    if (EVP_DigestSignInit(mdctx, NULL, NULL, NULL, pkey) <= 0) {
        fprintf(stderr, "Error initializing signature\n");
        ERR_print_errors_fp(stderr);
    }

    *signed_data_len = DRIP_UA_SIGNATURE_LEN;
    // Allocate memory for the signature
    if (EVP_DigestSign(mdctx, signed_data, signed_data_len, data, data_len) <= 0) {
        fprintf(stderr, "Error updating signature\n");
        ERR_print_errors_fp(stderr);
    }


    // Optional: Verify the signature to check if it's correct
    EVP_MD_CTX *verify_ctx = EVP_MD_CTX_new();
    if(verify_ctx == NULL) {
        fprintf(stderr, "Error initializing signature verification\n");
        ERR_print_errors_fp(stderr);
    }


    // Initialize verification operation
    EVP_DigestVerifyInit(verify_ctx, NULL, NULL, NULL, pkey);
    // Initialize verification operation
    if (1 != EVP_DigestVerifyInit(verify_ctx, NULL, NULL, NULL, pkey)) {
        fprintf(stderr, "Error verifying signature\n");
        ERR_print_errors_fp(stderr);
    }


    // Verify the signature
    if (1 == EVP_DigestVerify(verify_ctx, signed_data, *signed_data_len, data, data_len)) {
        printf("Signature verified successfully!\n");
    } else {
        printf("Signature verification failed!\n");
    }

    // Clean up
    EVP_MD_CTX_free(mdctx);
    EVP_MD_CTX_free(verify_ctx);
    EVP_PKEY_free(pkey);
    EVP_cleanup();
    ERR_free_strings();
}